package org.naokishibata.generic;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

public class KeySensor implements KeyListener, ComponentListener {
    private static Toolkit defaultToolkit = Toolkit.getDefaultToolkit();

    private long keyState = 0, keyState2 = 0;

    private LinkedList<Character> keyQueue = new LinkedList<Character>();
    private static final int keyQueueLength = 256;

    public final static long LEFT      = 1L << 0;
    public final static long DOWN      = 1L << 1;
    public final static long RIGHT     = 1L << 2;
    public final static long UP        = 1L << 3;
    public final static long PAGE_UP   = 1L << 4;
    public final static long PAGE_DOWN = 1L << 5;
    public final static long SHIFT     = 1L << 6;
    public final static long CONTROL   = 1L << 7;
    public final static long CTRL      = 1L << 7;
    public final static long TAB       = 1L << 8;
    public final static long ESC       = 1L << 9;
    public final static long ESCAPE    = 1L << 9;
    public final static long ALT       = 1L << 10;
    public final static long SPACE     = 1L << 11;
    public final static long ENTER     = 1L << 12;
    public final static long RETURN    = 1L << 12;
    public final static long HOME      = 1L << 13;
    public final static long END       = 1L << 14;
    public final static long INS       = 1L << 15;
    public final static long INSERT    = 1L << 15;
    public final static long DEL       = 1L << 16;
    public final static long DELETE    = 1L << 16;
    public final static long BS        = 1L << 17;
    public final static long BACKSPACE = 1L << 17;
    public final static long PAUSE     = 1L << 18;
    public final static long PRTSC     = 1L << 19;

    public final static long CAPSLOCK  = 1L << 20;
    public final static long NUMLOCK   = 1L << 21;
    public final static long SCRLOCK   = 1L << 22;

    public final static long F1        = 1L << 51;
    public final static long F2        = 1L << 52;
    public final static long F3        = 1L << 53;
    public final static long F4        = 1L << 54;
    public final static long F5        = 1L << 55;
    public final static long F6        = 1L << 56;
    public final static long F7        = 1L << 57;
    public final static long F8        = 1L << 58;
    public final static long F9        = 1L << 59;
    public final static long F10       = 1L << 60;
    public final static long F11       = 1L << 61;
    public final static long F12       = 1L << 62;


    public final static long ALNUM_0   = 1L << 0;
    public final static long ALNUM_1   = 1L << 1;
    public final static long ALNUM_2   = 1L << 2;
    public final static long ALNUM_3   = 1L << 3;
    public final static long ALNUM_4   = 1L << 4;
    public final static long ALNUM_5   = 1L << 5;
    public final static long ALNUM_6   = 1L << 6;
    public final static long ALNUM_7   = 1L << 7;
    public final static long ALNUM_8   = 1L << 8;
    public final static long ALNUM_9   = 1L << 9;

    public final static long ALNUM_A   = 1L << 10;
    public final static long ALNUM_B   = 1L << 11;
    public final static long ALNUM_C   = 1L << 12;
    public final static long ALNUM_D   = 1L << 13;
    public final static long ALNUM_E   = 1L << 14;
    public final static long ALNUM_F   = 1L << 15;
    public final static long ALNUM_G   = 1L << 16;
    public final static long ALNUM_H   = 1L << 17;
    public final static long ALNUM_I   = 1L << 18;
    public final static long ALNUM_J   = 1L << 19;
    public final static long ALNUM_K   = 1L << 20;
    public final static long ALNUM_L   = 1L << 21;
    public final static long ALNUM_M   = 1L << 22;
    public final static long ALNUM_N   = 1L << 23;
    public final static long ALNUM_O   = 1L << 24;
    public final static long ALNUM_P   = 1L << 25;
    public final static long ALNUM_Q   = 1L << 26;
    public final static long ALNUM_R   = 1L << 27;
    public final static long ALNUM_S   = 1L << 28;
    public final static long ALNUM_T   = 1L << 29;
    public final static long ALNUM_U   = 1L << 30;
    public final static long ALNUM_V   = 1L << 31;
    public final static long ALNUM_W   = 1L << 32;
    public final static long ALNUM_X   = 1L << 33;
    public final static long ALNUM_Y   = 1L << 34;
    public final static long ALNUM_Z   = 1L << 35;


    public long keyState() { return keyState; }
    public long keyState2() { return keyState2; }

    public char getKeyTyped() {
	char c;
	synchronized(keyQueue) {
	    if (keyQueue.size() == 0) {
		c = '\0';
	    } else {
		c = (keyQueue.poll());
	    }
	}
	return c;
    }

    public char peekKeyTyped() {
	char c;
	synchronized(keyQueue) {
	    if (keyQueue.size() == 0) {
		c = '\0';
	    } else {
		c = keyQueue.getFirst();
	    }
	}
	return c;
    }

    //

    public KeySensor() {}

    public KeySensor(Component c) {
	this();

	installOn(c);
    }

    public void installOn(Component c) {
	keyState = 0;
	keyState2 = 0;
	keyQueue.clear();

	c.addKeyListener(this);
	c.addComponentListener(this);
	c.setFocusable(true);
    }

    public void uninstallFrom(Component c) {
	c.removeKeyListener(this);
	c.removeComponentListener(this);
    }

    public void reset() {
	keyState = 0;
	keyState2 = 0;
	keyQueue.clear();
    }

    //

    public void keyTyped(KeyEvent e) {
	char c = e.getKeyChar();
	synchronized(keyQueue) {
	    keyQueue.offer(c);

	    while(keyQueue.size() > keyQueueLength) {
		keyQueue.poll();
	    }
	}
    }

    public void keyPressed(KeyEvent e) {
	switch(e.getKeyCode()) {
	case KeyEvent.VK_UP:
	    keyState |= UP;
	    break;
	case KeyEvent.VK_DOWN:
	    keyState |= DOWN;
	    break;
	case KeyEvent.VK_RIGHT:
	    keyState |= RIGHT;
	    break;
	case KeyEvent.VK_LEFT:
	    keyState |= LEFT;
	    break;
	case KeyEvent.VK_PAGE_UP:
	    keyState |= PAGE_UP;
	    break;
	case KeyEvent.VK_PAGE_DOWN:
	    keyState |= PAGE_DOWN;
	    break;
	case KeyEvent.VK_SHIFT:
	    keyState |= SHIFT;
	    break;
	case KeyEvent.VK_CONTROL:
	    keyState |= CONTROL;
	    break;
	case KeyEvent.VK_TAB:
	    keyState |= TAB;
	    break;
	case KeyEvent.VK_ESCAPE:
	    keyState |= ESCAPE;
	    break;
	case KeyEvent.VK_ALT:
	    keyState |= ALT;
	    break;
	case KeyEvent.VK_SPACE:
	    keyState |= SPACE;
	    break;
	case KeyEvent.VK_ENTER:
	    keyState |= ENTER;
	    break;
	case KeyEvent.VK_HOME:
	    keyState |= HOME;
	    break;
	case KeyEvent.VK_END:
	    keyState |= END;
	    break;

	case KeyEvent.VK_INSERT:
	    keyState |= INSERT;
	    break;
	case KeyEvent.VK_DELETE:
	    keyState |= DELETE;
	    break;
	case KeyEvent.VK_BACK_SPACE:
	    keyState |= BACKSPACE;
	    break;
	case KeyEvent.VK_PAUSE:
	    keyState |= PAUSE;
	    break;
	case KeyEvent.VK_PRINTSCREEN:
	    keyState |= PRTSC;
	    break;


	case KeyEvent.VK_CAPS_LOCK:
	    if (defaultToolkit.getLockingKeyState(KeyEvent.VK_CAPS_LOCK)) {
		keyState |= CAPSLOCK;
	    } else {
		keyState &= ~CAPSLOCK;
	    }
	    break;
	case KeyEvent.VK_NUM_LOCK:
	    if (defaultToolkit.getLockingKeyState(KeyEvent.VK_NUM_LOCK)) {
		keyState |= NUMLOCK;
	    } else {
		keyState &= ~NUMLOCK;
	    }
	    break;
	case KeyEvent.VK_SCROLL_LOCK:
	    if (defaultToolkit.getLockingKeyState(KeyEvent.VK_SCROLL_LOCK)) {
		keyState |= SCRLOCK;
	    } else {
		keyState &= ~SCRLOCK;
	    }
	    break;

	case KeyEvent.VK_F1:
	    keyState |= F1;
	    break;
	case KeyEvent.VK_F2:
	    keyState |= F2;
	    break;
	case KeyEvent.VK_F3:
	    keyState |= F3;
	    break;
	case KeyEvent.VK_F4:
	    keyState |= F4;
	    break;
	case KeyEvent.VK_F5:
	    keyState |= F5;
	    break;
	case KeyEvent.VK_F6:
	    keyState |= F6;
	    break;
	case KeyEvent.VK_F7:
	    keyState |= F7;
	    break;
	case KeyEvent.VK_F8:
	    keyState |= F8;
	    break;
	case KeyEvent.VK_F9:
	    keyState |= F9;
	    break;
	case KeyEvent.VK_F10:
	    keyState |= F10;
	    break;
	case KeyEvent.VK_F11:
	    keyState |= F11;
	    break;
	case KeyEvent.VK_F12:
	    keyState |= F12;
	    break;

	case KeyEvent.VK_0:
	    keyState2 |= ALNUM_0;
	    break;
	case KeyEvent.VK_1:
	    keyState2 |= ALNUM_1;
	    break;
	case KeyEvent.VK_2:
	    keyState2 |= ALNUM_2;
	    break;
	case KeyEvent.VK_3:
	    keyState2 |= ALNUM_3;
	    break;
	case KeyEvent.VK_4:
	    keyState2 |= ALNUM_4;
	    break;
	case KeyEvent.VK_5:
	    keyState2 |= ALNUM_5;
	    break;
	case KeyEvent.VK_6:
	    keyState2 |= ALNUM_6;
	    break;
	case KeyEvent.VK_7:
	    keyState2 |= ALNUM_7;
	    break;
	case KeyEvent.VK_8:
	    keyState2 |= ALNUM_8;
	    break;
	case KeyEvent.VK_9:
	    keyState2 |= ALNUM_9;
	    break;

	case KeyEvent.VK_A:
	    keyState2 |= ALNUM_A;
	    break;
	case KeyEvent.VK_B:
	    keyState2 |= ALNUM_B;
	    break;
	case KeyEvent.VK_C:
	    keyState2 |= ALNUM_C;
	    break;
	case KeyEvent.VK_D:
	    keyState2 |= ALNUM_D;
	    break;
	case KeyEvent.VK_E:
	    keyState2 |= ALNUM_E;
	    break;
	case KeyEvent.VK_F:
	    keyState2 |= ALNUM_F;
	    break;
	case KeyEvent.VK_G:
	    keyState2 |= ALNUM_G;
	    break;
	case KeyEvent.VK_H:
	    keyState2 |= ALNUM_H;
	    break;
	case KeyEvent.VK_I:
	    keyState2 |= ALNUM_I;
	    break;
	case KeyEvent.VK_J:
	    keyState2 |= ALNUM_J;
	    break;
	case KeyEvent.VK_K:
	    keyState2 |= ALNUM_K;
	    break;
	case KeyEvent.VK_L:
	    keyState2 |= ALNUM_L;
	    break;
	case KeyEvent.VK_M:
	    keyState2 |= ALNUM_M;
	    break;
	case KeyEvent.VK_N:
	    keyState2 |= ALNUM_N;
	    break;
	case KeyEvent.VK_O:
	    keyState2 |= ALNUM_O;
	    break;
	case KeyEvent.VK_P:
	    keyState2 |= ALNUM_P;
	    break;
	case KeyEvent.VK_Q:
	    keyState2 |= ALNUM_Q;
	    break;
	case KeyEvent.VK_R:
	    keyState2 |= ALNUM_R;
	    break;
	case KeyEvent.VK_S:
	    keyState2 |= ALNUM_S;
	    break;
	case KeyEvent.VK_T:
	    keyState2 |= ALNUM_T;
	    break;
	case KeyEvent.VK_U:
	    keyState2 |= ALNUM_U;
	    break;
	case KeyEvent.VK_V:
	    keyState2 |= ALNUM_V;
	    break;
	case KeyEvent.VK_W:
	    keyState2 |= ALNUM_W;
	    break;
	case KeyEvent.VK_X:
	    keyState2 |= ALNUM_X;
	    break;
	case KeyEvent.VK_Y:
	    keyState2 |= ALNUM_Y;
	    break;
	case KeyEvent.VK_Z:
	    keyState2 |= ALNUM_Z;
	    break;
	}
    }

    public void keyReleased(KeyEvent e) {
	switch(e.getKeyCode()) {
	case KeyEvent.VK_UP:
	    keyState &= ~UP;
	    break;
	case KeyEvent.VK_DOWN:
	    keyState &= ~DOWN;
	    break;
	case KeyEvent.VK_RIGHT:
	    keyState &= ~RIGHT;
	    break;
	case KeyEvent.VK_LEFT:
	    keyState &= ~LEFT;
	    break;
	case KeyEvent.VK_PAGE_UP:
	    keyState &= ~PAGE_UP;
	    break;
	case KeyEvent.VK_PAGE_DOWN:
	    keyState &= ~PAGE_DOWN;
	    break;
	case KeyEvent.VK_SHIFT:
	    keyState &= ~SHIFT;
	    break;
	case KeyEvent.VK_CONTROL:
	    keyState &= ~CONTROL;
	    break;
	case KeyEvent.VK_TAB:
	    keyState &= ~TAB;
	    break;
	case KeyEvent.VK_ESCAPE:
	    keyState &= ~ESCAPE;
	    break;
	case KeyEvent.VK_ALT:
	    keyState &= ~ALT;
	    break;
	case KeyEvent.VK_SPACE:
	    keyState &= ~SPACE;
	    break;
	case KeyEvent.VK_ENTER:
	    keyState &= ~ENTER;
	    break;
	case KeyEvent.VK_HOME:
	    keyState &= ~HOME;
	    break;
	case KeyEvent.VK_END:
	    keyState &= ~END;
	    break;

	case KeyEvent.VK_INSERT:
	    keyState &= ~INSERT;
	    break;
	case KeyEvent.VK_DELETE:
	    keyState &= ~DELETE;
	    break;
	case KeyEvent.VK_BACK_SPACE:
	    keyState &= ~BACKSPACE;
	    break;
	case KeyEvent.VK_PAUSE:
	    keyState &= ~PAUSE;
	    break;
	case KeyEvent.VK_PRINTSCREEN:
	    keyState &= ~PRTSC;
	    break;

	case KeyEvent.VK_CAPS_LOCK:
	    if (defaultToolkit.getLockingKeyState(KeyEvent.VK_CAPS_LOCK)) {
		keyState |= CAPSLOCK;
	    } else {
		keyState &= ~CAPSLOCK;
	    }
	    break;
	case KeyEvent.VK_NUM_LOCK:
	    if (defaultToolkit.getLockingKeyState(KeyEvent.VK_NUM_LOCK)) {
		keyState |= NUMLOCK;
	    } else {
		keyState &= ~NUMLOCK;
	    }
	    break;
	case KeyEvent.VK_SCROLL_LOCK:
	    if (defaultToolkit.getLockingKeyState(KeyEvent.VK_SCROLL_LOCK)) {
		keyState |= SCRLOCK;
	    } else {
		keyState &= ~SCRLOCK;
	    }
	    break;

	case KeyEvent.VK_F1:
	    keyState &= ~F1;
	    break;
	case KeyEvent.VK_F2:
	    keyState &= ~F2;
	    break;
	case KeyEvent.VK_F3:
	    keyState &= ~F3;
	    break;
	case KeyEvent.VK_F4:
	    keyState &= ~F4;
	    break;
	case KeyEvent.VK_F5:
	    keyState &= ~F5;
	    break;
	case KeyEvent.VK_F6:
	    keyState &= ~F6;
	    break;
	case KeyEvent.VK_F7:
	    keyState &= ~F7;
	    break;
	case KeyEvent.VK_F8:
	    keyState &= ~F8;
	    break;
	case KeyEvent.VK_F9:
	    keyState &= ~F9;
	    break;
	case KeyEvent.VK_F10:
	    keyState &= ~F10;
	    break;
	case KeyEvent.VK_F11:
	    keyState &= ~F11;
	    break;
	case KeyEvent.VK_F12:
	    keyState &= ~F12;
	    break;

	case KeyEvent.VK_0:
	    keyState2 &= ~ ALNUM_0;
	    break;
	case KeyEvent.VK_1:
	    keyState2 &= ~ ALNUM_1;
	    break;
	case KeyEvent.VK_2:
	    keyState2 &= ~ ALNUM_2;
	    break;
	case KeyEvent.VK_3:
	    keyState2 &= ~ ALNUM_3;
	    break;
	case KeyEvent.VK_4:
	    keyState2 &= ~ ALNUM_4;
	    break;
	case KeyEvent.VK_5:
	    keyState2 &= ~ ALNUM_5;
	    break;
	case KeyEvent.VK_6:
	    keyState2 &= ~ ALNUM_6;
	    break;
	case KeyEvent.VK_7:
	    keyState2 &= ~ ALNUM_7;
	    break;
	case KeyEvent.VK_8:
	    keyState2 &= ~ ALNUM_8;
	    break;
	case KeyEvent.VK_9:
	    keyState2 &= ~ ALNUM_9;
	    break;

	case KeyEvent.VK_A:
	    keyState2 &= ~ ALNUM_A;
	    break;
	case KeyEvent.VK_B:
	    keyState2 &= ~ ALNUM_B;
	    break;
	case KeyEvent.VK_C:
	    keyState2 &= ~ ALNUM_C;
	    break;
	case KeyEvent.VK_D:
	    keyState2 &= ~ ALNUM_D;
	    break;
	case KeyEvent.VK_E:
	    keyState2 &= ~ ALNUM_E;
	    break;
	case KeyEvent.VK_F:
	    keyState2 &= ~ ALNUM_F;
	    break;
	case KeyEvent.VK_G:
	    keyState2 &= ~ ALNUM_G;
	    break;
	case KeyEvent.VK_H:
	    keyState2 &= ~ ALNUM_H;
	    break;
	case KeyEvent.VK_I:
	    keyState2 &= ~ ALNUM_I;
	    break;
	case KeyEvent.VK_J:
	    keyState2 &= ~ ALNUM_J;
	    break;
	case KeyEvent.VK_K:
	    keyState2 &= ~ ALNUM_K;
	    break;
	case KeyEvent.VK_L:
	    keyState2 &= ~ ALNUM_L;
	    break;
	case KeyEvent.VK_M:
	    keyState2 &= ~ ALNUM_M;
	    break;
	case KeyEvent.VK_N:
	    keyState2 &= ~ ALNUM_N;
	    break;
	case KeyEvent.VK_O:
	    keyState2 &= ~ ALNUM_O;
	    break;
	case KeyEvent.VK_P:
	    keyState2 &= ~ ALNUM_P;
	    break;
	case KeyEvent.VK_Q:
	    keyState2 &= ~ ALNUM_Q;
	    break;
	case KeyEvent.VK_R:
	    keyState2 &= ~ ALNUM_R;
	    break;
	case KeyEvent.VK_S:
	    keyState2 &= ~ ALNUM_S;
	    break;
	case KeyEvent.VK_T:
	    keyState2 &= ~ ALNUM_T;
	    break;
	case KeyEvent.VK_U:
	    keyState2 &= ~ ALNUM_U;
	    break;
	case KeyEvent.VK_V:
	    keyState2 &= ~ ALNUM_V;
	    break;
	case KeyEvent.VK_W:
	    keyState2 &= ~ ALNUM_W;
	    break;
	case KeyEvent.VK_X:
	    keyState2 &= ~ ALNUM_X;
	    break;
	case KeyEvent.VK_Y:
	    keyState2 &= ~ ALNUM_Y;
	    break;
	case KeyEvent.VK_Z:
	    keyState2 &= ~ ALNUM_Z;
	    break;
	}
    }

    // ComponentListener

    public void componentHidden(ComponentEvent e) {
	keyState = 0;
	keyState2 = 0;
	keyQueue.clear();
    }

    public void componentMoved(ComponentEvent e) {
    }

    public void componentResized(ComponentEvent e) {
    }

    public void componentShown(ComponentEvent e) {
	keyState = 0;
	keyState2 = 0;
	keyQueue.clear();
    }
}
