package org.naokishibata.generic;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.awt.geom.*;
import javax.imageio.*;
import javax.swing.*;
import javax.swing.text.*;
import java.util.*;
import java.io.*;

public class Scribble2D {
    AnimationPanel canvas;
    MouseSensor mouseSensor;
    KeySensor keySensor;

    private JFrame frame;
    private Color bgColor = Color.white;
    private VolatileImage vImg;
    private Graphics2D g2d;
    private int width, height;

    public Scribble2D(int width, int height, boolean ebs) {
	this.width = width;
	this.height = height;
	canvas = new AnimationPanel(width, height, ebs);
	canvas.setBackgroundColor(bgColor);
	vImg = canvas.newVolatileImage(width, height);
	g2d = vImg.createGraphics();
    }

    public Scribble2D(int width, int height) { this(width, height, false); }

    public Scribble2D() { this(640, 480); }

    public JPanel makePanel() {
	JPanel c = canvas.makePanel();

	if (mouseSensor == null) mouseSensor = new MouseSensor();
	mouseSensor.installOn(c);
	if (keySensor == null) keySensor = new KeySensor();
	keySensor.installOn(c);

	AWTUtils.makeFocusable(c);

	return c;
    }

    public JFrame makeFrame(String name) {
	if (frame != null) return frame;

        frame = new JFrame(name);

	Insets in = frame.getInsets();
	frame.setSize(width + in.left + in.right, height + in.top  + in.bottom);

        frame.setResizable(true);
        ((JFrame)frame).setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);

        ((JFrame)frame).getContentPane().add("Center", makePanel());

        AWTUtils.placeAtTheCenter(frame);

        frame.pack();
        frame.setVisible(true);

        while(frame.getWidth() <= 200 && frame.getHeight() <= 100) {
            in = frame.getInsets();
            frame.setSize(width + in.left + in.right, height + in.top  + in.bottom);

            try { Thread.sleep(100); } catch(InterruptedException e) {}
        }

	//System.out.println(frame.getWidth() + ", " + frame.getHeight());

        AWTUtils.placeAtTheCenter(frame);

	return frame;
    }

    public JFrame makeFrame() { return makeFrame(MiscUtils.getMainClassName()); }

    public DisplayMode[] getDisplayModes() { return canvas.getDisplayModes(); }

    /*
    public void switchToFullScreen() {
	Frame f = canvas.makeFullScreenFrame();

	if (mouseSensor == null) mouseSensor = new MouseSensor();
	mouseSensor.installOn(f);
	if (keySensor == null) keySensor = new KeySensor();
	keySensor.installOn(f);
    }

    public void backFromFullScreen() { canvas.setVisible(false); }
    */

    public void render(Graphics g) {
	canvas.render(g);
    }

    //

    public void saveImage(String fmt, File f) throws IOException {
	BufferedImage bImg;

	bImg = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
	Graphics2D newg2d = bImg.createGraphics();

	newg2d.drawImage(vImg, 0, 0, null);

	ImageIO.write(bImg, fmt, f);

	newg2d.dispose();
    }

    //

    public void setVisible(boolean b) { canvas.setVisible(b); }
    public void setIconifyWhenClosed(boolean b) { canvas.setIconifyWhenClosed(b); }
    public void setExitWhenClosed(boolean b) { canvas.setExitWhenClosed(b); }

    public void setBackgroundColor(Color c) {
	bgColor = c;
	canvas.setBackgroundColor(c);
    }

    //

    private boolean useAntialias = true;
    private boolean freezed = false;
    private Object freezeLock = new Object();

    private Object lock = new Object();

    public void enableAntiAlias(boolean b) { 
	useAntialias = b;
	if (b) {
	    g2d.addRenderingHints(new RenderingHints(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON));
	} else {
	    g2d.addRenderingHints(new RenderingHints(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF));
	}
    }

    public void enableAntialias(boolean b) { 
	useAntialias = b;
	if (b) {
	    g2d.addRenderingHints(new RenderingHints(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON));
	} else {
	    g2d.addRenderingHints(new RenderingHints(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF));
	}
    }

    public void setFreezed(boolean b) {
	freezed = b;
	if (!freezed) {
	    updateCanvas(0);
	    synchronized(freezeLock) {
		freezeLock.notify();
	    }
	}
    }

    //

    private long lastUpdate = 0;

    private void updateCanvas(long u) {
	if (width != canvas.getWidth() || height != canvas.getHeight()) {
	    width = canvas.getWidth();
	    height = canvas.getHeight();

	    VolatileImage newImg = canvas.newVolatileImage(width, height);
	    Graphics2D newg2d = newImg.createGraphics();

	    newg2d.setColor(bgColor);
	    newg2d.fillRect(0, 0, width, height);
	    newg2d.drawImage(vImg, 0, 0, null);

	    g2d.dispose();

	    g2d = newg2d;
	    vImg = newImg;
	}

	if (freezed) return;

	long t = System.currentTimeMillis();
	long s = lastUpdate + u;

	if (t < s) {
	    try { Thread.sleep(s-t); } catch(InterruptedException e) {}
	    lastUpdate = s;
	} else if (t < s + u*3) {
	    lastUpdate = s;
	} else {
	    lastUpdate = t;
	}

	Graphics2D g = canvas.getGraphics2D();
	g.drawImage(vImg, 0, 0, null);
	canvas.swapBuffer();
    }

    public void showFrame(long t) {
	freezed = false;
	updateCanvas(t);
	synchronized(freezeLock) {
	    freezeLock.notify();
	}
	freezed = true;
    }

    //

    public void wipe() {
	synchronized(lock) {
	    g2d.setColor(bgColor);
	    g2d.fillRect(0, 0, canvas.getWidth(), canvas.getHeight());

	    updateCanvas(0);
	}
    }

    public void drawImage(double x, double y, Image img) {
	synchronized(lock) {
	    g2d.drawImage(img, (int)x, (int)y, null);

	    updateCanvas(0);
	}
    }

    public void drawImage(double x0, double y0, double x1, double y1, Image img) {
	synchronized(lock) {
	    g2d.drawImage(img, (int)Math.rint(x0), (int)Math.rint(y0), (int)Math.rint(x1), (int)Math.rint(y1), 
			  0, 0, img.getWidth(null)-1, img.getHeight(null)-1, null);

	    updateCanvas(0);
	}
    }

    public void drawLine(double x0, double y0, double x1, double y1, double thickness, Color c) {
	GeneralPath gp = new GeneralPath();

	gp.moveTo((float)x0, (float)y0);
	gp.lineTo((float)x1, (float)y1);

	synchronized(lock) {
	    g2d.setColor(c);
	    g2d.setStroke(new BasicStroke((float)thickness, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL));
	    g2d.draw(gp);

	    updateCanvas(0);
	}
    }

    public void drawLines(double thickness, Color c, double ... coords) {
	if (coords.length < 4) return;

	GeneralPath gp = new GeneralPath();

	gp.moveTo((float)coords[0], (float)coords[1]);

	for(int i=2;i<coords.length;i+=2) {
	    gp.lineTo((float)coords[i], (float)coords[i+1]);
	}

	//gp.closePath();

	synchronized(lock) {
	    g2d.setColor(c);
	    g2d.setStroke(new BasicStroke((float)thickness, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL));
	    g2d.draw(gp);

	    updateCanvas(0);
	}
    }

    public void fillLines(Color c, double ... coords) {
	if (coords.length < 4) return;

	GeneralPath gp = new GeneralPath();

	gp.moveTo((float)coords[0], (float)coords[1]);

	for(int i=2;i<coords.length;i+=2) {
	    gp.lineTo((float)coords[i], (float)coords[i+1]);
	}

	gp.closePath();

	synchronized(lock) {
	    g2d.setColor(c);
	    g2d.fill(gp);

	    updateCanvas(0);
	}
    }

    public void drawBox(double x0, double y0, double x1, double y1, double thickness, Color c) {
	GeneralPath gp = new GeneralPath();

	gp.moveTo((float)x0, (float)y0);
	gp.lineTo((float)x1, (float)y0);
	gp.lineTo((float)x1, (float)y1);
	gp.lineTo((float)x0, (float)y1);
	gp.closePath();

	synchronized(lock) {
	    g2d.setColor(c);
	    g2d.setStroke(new BasicStroke((float)thickness, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL));
	    g2d.draw(gp);

	    updateCanvas(0);
	}
    }

    public void fillBox(double x0, double y0, double x1, double y1, Color c) {
	GeneralPath gp = new GeneralPath();

	gp.moveTo((float)x0, (float)y0);
	gp.lineTo((float)x1, (float)y0);
	gp.lineTo((float)x1, (float)y1);
	gp.lineTo((float)x0, (float)y1);
	gp.closePath();

	synchronized(lock) {
	    g2d.setColor(c);
	    g2d.fill(gp);

	    updateCanvas(0);
	}
    }

    public void fillRoundRect(double x0, double y0, double x1, double y1, double r, Color c) {
	synchronized(lock) {
	    g2d.setColor(c);
	    g2d.fillRoundRect((int)x0, (int)y0, (int)(x1-x0+1), (int)(y1-y0+1), (int)r, (int)r);

	    updateCanvas(0);
	}
    }

    public void drawCircle(double x, double y, double r, double thickness, Color c) {
	Arc2D.Double a = new Arc2D.Double(x-r, y-r, r*2, r*2, 0, 360, Arc2D.OPEN);

	synchronized(lock) {
	    g2d.setColor(c);
	    g2d.setStroke(new BasicStroke((float)thickness, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL));
	    g2d.draw(a);

	    updateCanvas(0);
	}
    }

    public void fillCircle(double x, double y, double r, Color c) {
	Arc2D.Double a = new Arc2D.Double(x-r, y-r, r*2, r*2, 0, 360, Arc2D.OPEN);

	synchronized(lock) {
	    g2d.setColor(c);
	    g2d.fill(a);

	    updateCanvas(0);
	}
    }

    public void fillArc(double x, double y, double r, double startAngle, double arcAngle, Color c) {
	Arc2D.Double a = new Arc2D.Double(x-r, y-r, r*2, r*2, startAngle*180.0/Math.PI, arcAngle*180.0/Math.PI, Arc2D.PIE);

	synchronized(lock) {
	    g2d.setColor(c);
	    g2d.fill(a);

	    updateCanvas(0);
	}
    }

    public static final int BOTTOM_LEFT = 0;
    public static final int BOTTOM_CENTER = 1;
    public static final int BOTTOM_RIGHT = 2;
    public static final int LEFT = 3;
    public static final int CENTER = 4;
    public static final int RIGHT = 5;
    public static final int TOP_LEFT = 6;
    public static final int TOP_CENTER = 7;
    public static final int TOP_RIGHT = 8;

    public void drawString(String str, double x, double y, int pos, int size, Color c) {
	Font f = new Font("sansserif", Font.PLAIN, size);
	Rectangle2D r = f.getStringBounds(str, g2d.getFontRenderContext());

	if (pos < 0 || pos > 8) pos = 4;

	double rx = r.getX() + (pos % 3) * r.getWidth()/2;
	double ry = r.getY() + (2 - pos / 3) * r.getHeight()/2;

	x -= rx;
	y -= ry;

	synchronized(lock) {
	    g2d.setFont(f);
	    g2d.setColor(c);
	    g2d.drawString(str, (float)x, (float)y);

	    updateCanvas(0);
	}
    }

    public double inputDouble(String message, double def) {
	Double d = StringInputDialog.showDialogForDoubleInput(frame, message, "Input a double value", StringInputDialog.OK_OPTION, def);

	if (d == null) return def;

	return d.doubleValue();
    }

    public double inputDouble(String message) { return inputDouble(message, 0); }

    public int inputInt(String message, int def) {
	Integer i = StringInputDialog.showDialogForIntInput(frame, message, "Input an integer value", StringInputDialog.OK_OPTION, def);

	if (i == null) return def;

	return i.intValue();
    }

    public int inputInt(String message) { return inputInt(message, 0); }

    public String inputString(String message, String def) {
	String s = StringInputDialog.showDialogForStringInput(frame, message, "Input a string", StringInputDialog.OK_OPTION, def);

	if (s == null) return def;

	return s;
    }

    public String inputString(String message) { return inputString(message, ""); }

    public void showMessage(String message) {
	JOptionPane.showMessageDialog(frame, message, "Message", JOptionPane.PLAIN_MESSAGE);
    }

    //

    public int mouseX() { return mouseSensor.posX(); }
    public int mouseY() { return mouseSensor.posY(); }
    public int mouseButton() { return mouseSensor.button(); }
    public int mouseWheel() { return mouseSensor.wheel(); }
    public long keyState() { return keySensor.keyState(); }
    public int width() { return width; }
    public int height() { return height; }
}
