package org.naokishibata.examples;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.nio.*;
import java.util.*;

import com.sun.opengl.util.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;

import org.naokishibata.generic.*;
import org.naokishibata.math.*;
import org.naokishibata.ocvimgio.*;
import org.naokishibata.armarker.*;

public class AccuTest implements GLEventListener, WindowListener, Runnable {
    static final boolean debug = true;
    final int MAPSIZE;

    private GLCanvas canvas;
    private GLU glu;
    private GLUT glut;
    private int texName;
    private double tw, th, near, tright, ttop, tnear;
    private int lastWidth, lastHeight;

    private KeySensor ks;

    private int[] viewport = new int[4];
    private double[] modelMatrix = new double[16];
    private double[] projMatrix = new double[16];

    CvCapture capture;
    int iw, ih;
    ByteBuffer bb;

    AccuTag st;

    private boolean firstShow = true;

    double aovDeg, camAspect, camDistortion;
    double aov;

    String fnout;
    IplImage oimg = null;
    BGVideoWriter vw = null;

    public AccuTest(String fn, String fnout, int iw0, int ih0, double aovDeg, double camAspect, double camDistortion, boolean enableSubpixel, int clDeviceID) throws Exception {
	this.aovDeg = aovDeg;
	aov = aovDeg / 180.0 * Math.PI;

	this.camAspect = camAspect;
	this.camDistortion = camDistortion;

	if (fn == null) {
	    capture = CvCapture.cvCreateCameraCapture(0);
	    capture.cvSetCaptureProperty(CvCapture.CV_CAP_PROP_FRAME_WIDTH , iw0);
	    capture.cvSetCaptureProperty(CvCapture.CV_CAP_PROP_FRAME_HEIGHT, ih0);
	} else {
	    capture = CvCapture.cvCreateFileCapture(fn);
	}

	IplImage iimg = capture.cvQueryFrame();

	iw = iimg.width();
	ih = iimg.height();

	this.fnout = fnout;
	if (fnout != null) {
	    vw = BGVideoWriter.cvCreateVideoWriter(fnout, "PIM1", 30, iw, ih, 1);
	    oimg = new IplImage(iw, ih, 8, 3);
	    Runtime.getRuntime().addShutdownHook(new Thread(this));
	}

	int shift = 11;
	MAPSIZE = 1 << shift;

	bb = ByteBuffer.allocateDirect(MAPSIZE * MAPSIZE * 3);

	st = new AccuTag(shift, iw, ih, clDeviceID);
	st.enableSubpixel = enableSubpixel;

	tw = (double)iw / MAPSIZE;
	th = (double)ih / MAPSIZE;

	//

        GLCapabilities caps = new GLCapabilities();

	caps.setAlphaBits(8);
	caps.setDepthBits(24);

        caps.setDoubleBuffered(true);
        caps.setHardwareAccelerated(true);

        canvas = new GLCanvas(caps, null, null, null);

        canvas.addGLEventListener(this);

	glu = new GLU();
	glut = new GLUT();

	ks = new KeySensor(canvas);
    }

    public void run() {
	if (vw != null) vw.close();
    }

    public void init(GLAutoDrawable drawable) {
	if (firstShow) {
	    firstShow = false;
	    if (debug) drawable.setGL(new DebugGL(drawable.getGL()));

	    GL gl = drawable.getGL();

	    gl.glMatrixMode(GL.GL_MODELVIEW);
	    gl.glLoadIdentity();

	    gl.glMatrixMode(GL.GL_PROJECTION);
	    gl.glLoadIdentity();
	    gl.glOrtho(0, 1, 0, 1, -1, 1);

	    //

	    int[] tmp = new int[1];
	    gl.glGenTextures(1, tmp, 0);
	    texName = tmp[0];

	    gl.glBindTexture(GL.GL_TEXTURE_2D, texName);
	    gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_GENERATE_MIPMAP, GL.GL_FALSE);
	    gl.glTexImage2D(GL.GL_TEXTURE_2D, 0, GL.GL_RGBA8, MAPSIZE, MAPSIZE, 0, GL.GL_BGR, GL.GL_UNSIGNED_BYTE, bb.rewind());

	    gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP);
	    gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, GL.GL_CLAMP);
	    gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST);
	    gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST);
	    gl.glTexEnvf(GL.GL_TEXTURE_ENV, GL.GL_TEXTURE_ENV_MODE, GL.GL_REPLACE);
	}
    }

    public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {
	reshapeMain(drawable.getGL(), width, height);
	lastWidth = width;
	lastHeight = height;
    }

    public void reshapeMain(GL gl, int width, int height) {
	double right, top;

	if ((double)iw / ih > (double)width / height) {
	    near   = ((double)iw / 2) / Math.tan(aov / 2);
	    top    = (double)height / width * iw / 2;
	    right  = (double)iw / 2;

	    ttop   = (double)ih / 2;
	    tright = (double)iw / 2;
	    tnear  = near;
	} else {
	    near   = ((double)iw / 2) / Math.tan(aov / 2);
	    top    = (double)ih / 2;
	    right  = (double)width / height * ih / 2;

	    ttop   = (double)ih / 2;
	    tright = (double)iw / 2;
	    tnear  = near;
	}

	System.out.println("width = " + width + ", height = " + height);
	System.out.println("iw = " + iw + ", ih = " + ih);
	System.out.println("top = " + top + ", right = " + right + ", near = " + near);

	top /= near;
	right /= near;
	near = 1;

	gl.glMatrixMode(GL.GL_PROJECTION);
	gl.glLoadIdentity();
	gl.glFrustum(-right, right, top, -top, 1, 10000);

	//

	gl.glViewport(0, 0, width, height);

	//

	gl.glGetIntegerv(GL.GL_VIEWPORT, viewport, 0);
	gl.glGetDoublev(GL.GL_MODELVIEW_MATRIX, modelMatrix, 0);
	gl.glGetDoublev(GL.GL_PROJECTION_MATRIX, projMatrix, 0);
    }

    public void displayChanged(GLAutoDrawable drawable, boolean modeChanged, boolean deviceChanged) {
    }

    public void display(GLAutoDrawable drawable) {
	IplImage iimg = capture.cvRetrieveFrame();
	capture.cvGrabFrame();

	LinkedList<AccuTag.Marker> list = st.recognize(iimg, aov, camAspect, camDistortion);

	//

	GL gl = drawable.getGL();

	gl.glClear(GL.GL_DEPTH_BUFFER_BIT);

	gl.glBindTexture(GL.GL_TEXTURE_2D, texName);

	gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_GENERATE_MIPMAP, GL.GL_FALSE);

	gl.glTexSubImage2D(GL.GL_TEXTURE_2D, 0, 0, 0, iw, ih, GL.GL_BGR, GL.GL_UNSIGNED_BYTE, iimg.getDirect().rewind());

	gl.glDepthMask( false );
	gl.glDisable( GL.GL_DEPTH_TEST );
	gl.glEnable(GL.GL_TEXTURE_2D);

	gl.glBegin( GL.GL_QUADS );
	gl.glTexCoord2d(0.0,th ); gl.glVertex3d(-tright,+ttop,-tnear);
	gl.glTexCoord2d(tw ,th ); gl.glVertex3d(+tright,+ttop,-tnear);
	gl.glTexCoord2d(tw ,0.0); gl.glVertex3d(+tright,-ttop,-tnear);
	gl.glTexCoord2d(0.0,0.0); gl.glVertex3d(-tright,-ttop,-tnear);
	gl.glEnd();

	gl.glDisable(GL.GL_TEXTURE_2D);
	gl.glEnable( GL.GL_DEPTH_TEST );
	gl.glDepthMask( true );

	for(AccuTag.Marker m : list) {
	    Vector3D v;

	    Vector3D vecx = m.x.minus(m.o);
	    Vector3D vecy = m.y.minus(m.o);
	    Vector3D vecz = vecx.cross(vecy).normalized();

	    Vector3D[] v3 = { m.o, m.x, m.x.plus(vecy), m.y };

	    gl.glLineWidth(1.0f);
	    gl.glColor3f(0.0f, 1.0f, 1.0f);

	    gl.glBegin(GL.GL_LINES);

	    for(int i=0;i<4;i++) {
		Vector3D v0, v1;

		v0 = v3[i]; v1 = v3[(i+1)&3];
		gl.glVertex3d(v0.x, v0.y, -v0.z);
		gl.glVertex3d(v1.x, v1.y, -v1.z);

		v0 = v3[i]; v1 = v3[i].plus(vecz);
		gl.glVertex3d(v0.x, v0.y, -v0.z);
		gl.glVertex3d(v1.x, v1.y, -v1.z);

		v0 = v3[i].plus(vecz); v1 = v3[(i+1)&3].plus(vecz);
		gl.glVertex3d(v0.x, v0.y, -v0.z);
		gl.glVertex3d(v1.x, v1.y, -v1.z);
	    }

	    gl.glEnd();

	    gl.glLineWidth(4.0f);
	    gl.glColor3f(0.0f, 0.0f, 1.0f);

	    gl.glBegin(GL.GL_LINES);

	    {
		Vector3D v0, v1;

		v0 = v3[0]; v1 = v3[0].plus(vecz);
		gl.glVertex3d(v0.x, v0.y, -v0.z);
		gl.glVertex3d(v1.x, v1.y, -v1.z);
	    }

	    gl.glEnd();

	    v = Vector3D.midpoint(m.x, m.y);

	    gl.glColor3f(0.8f, 0.0f, 0.0f);

	    drawStringCenter(gl, Vector3D.midpoint(m.x, m.y), "" + m.payload + " : " + ((int)Math.rint(100*v.length())) + " : " + ((int)Math.rint(Math.acos(-vecz.z)/Math.PI*180.0)));
	}

	if (vw != null) {
	    gl.glReadBuffer(GL.GL_FRONT);
	    gl.glReadPixels(0, 0, iw, ih, GL.GL_BGR, GL.GL_UNSIGNED_BYTE, oimg.getDirect().rewind());
	    gl.glFinish();
	    CvUtil.cvFlip(oimg, oimg, 0);
	    vw.cvWriteFrame(oimg);
	}
    }

    void drawStringCenter(GL gl, Vector3D pos, String s) {
	double[] winpos = new double[3];
	glu.gluProject(pos.x, pos.y, -pos.z, modelMatrix, 0, projMatrix, 0, viewport, 0, winpos, 0);

	gl.glMatrixMode(GL.GL_MODELVIEW);
	gl.glPushMatrix();
	gl.glLoadIdentity();

	gl.glMatrixMode(GL.GL_PROJECTION);
	gl.glPushMatrix();
	gl.glLoadIdentity();

	gl.glRasterPos2d((winpos[0] - s.length()*9/2) * 2 / (double)viewport[2] - 1, (winpos[1]-7) * 2 / (double)viewport[3] - 1);
	gl.glDisable( GL.GL_DEPTH_TEST );
	//glut.glutBitmapString(GLUT.BITMAP_9_BY_15, s);
	glut.glutBitmapString(GLUT.BITMAP_TIMES_ROMAN_24, s);
	gl.glEnable( GL.GL_DEPTH_TEST );

	gl.glPopMatrix();
	gl.glMatrixMode(GL.GL_MODELVIEW);
	gl.glPopMatrix();
    }

    public static void main(String[] args) throws Exception {
	// logitech C920, 1600x896
	//double aovDeg = 67.21;
	//double camAspect = 1.0217, camDistortion = 0.00;

	double aovDeg = 51.9;
	double camAspect = 1.0, camDistortion = 0.00;

	HashMap<String, Integer> options = new HashMap<String, Integer>();
	options.put("--size", 2);
	options.put("--aovdeg", 1);
	options.put("--aspect", 1);
	options.put("--distortion", 1);
	options.put("--nosubpixel", 0);
	options.put("--cldevice", 1);
	options.put("--videoin", 1);
	options.put("--videoout", 1);

	HashMap<String, String[]> cl = MiscUtils.analyseCommandLine(args, options);

	if (cl == null) {
	    System.out.println("usage java org.naokishibata.examples.AccuTest <options>");
	    System.out.println("options :");
	    System.out.println("  --size <width> <height>  specify captured image size");
	    System.out.println("  --aovdeg <degree>        specify angle of view in degree");
	    System.out.println("  --aspect <ratio>         specify aspect ratio of pixels");
	    System.out.println("  --distortion <number>    specify camera distortion");
	    System.out.println("  --nosubpixel             disable subpixel corner detection");
	    System.out.println("  --cldevice <number>      specify OpenCL device");
	    System.out.println("  --videoin <filename>     specify video file name for input");
	    System.out.println("  --videoout <filename>    specify video file name for output");
	    System.exit(-1);
	}

	String fn = null;
	int width = 1024, height = 768;
	boolean enableSubpixel = true;
	int clDeviceID = 0;
	String fnout = null;

	for(String k : cl.keySet()) {
	    String[] a = cl.get(k);
	    if (k.equals("--aovdeg")) {
		aovDeg = Double.parseDouble(a[0]);
	    } else if (k.equals("--aspect")) {
		camAspect = Double.parseDouble(a[0]);
	    } else if (k.equals("--distortion")) {
		camDistortion = Double.parseDouble(a[0]);
	    } else if (k.equals("--videoin")) {
		fn = a[0];
	    } else if (k.equals("--size")) {
		width = Integer.parseInt(a[0]);
		height = Integer.parseInt(a[1]);
	    } else if (k.equals("--nosubpixel")) {
		enableSubpixel = false;
	    } else if (k.equals("--cldevice")) {
		clDeviceID = Integer.parseInt(a[0]);
	    } else if (k.equals("--videoout")) {
		fnout = a[0];
	    }
	}

	JFrame frame = new JFrame();

	frame.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);

        frame.setLayout(new BorderLayout());

        frame.setSize(width, height);

	AWTUtils.placeAtTheCenter(frame);

	AccuTest instance = new AccuTest(fn, fnout, width, height, aovDeg, camAspect, camDistortion, enableSubpixel, clDeviceID);

	frame.addWindowListener(instance);

        frame.setVisible(true);

	AWTUtils.setContentSize(frame, new Dimension(width, height));

        frame.add(instance.canvas, BorderLayout.CENTER);

        //frame.setVisible(true);

	FPSAnimator anim = new FPSAnimator(instance.canvas, 60);
	anim.start();
    }

    public void windowActivated(WindowEvent e) {}
    public void windowClosed(WindowEvent e) {}
    public void windowDeactivated(WindowEvent e) {}
    public void windowDeiconified(WindowEvent e) {}
    public void windowIconified(WindowEvent e) {}
    public void windowOpened(WindowEvent e) {}

    public void windowClosing(WindowEvent e) {
	BGVideoWriter vw2 = vw;
	vw = null;
	if (vw2 != null) vw2.close();

	System.exit(0);
    }
}
