package org.naokishibata.generic;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.awt.geom.*;
import javax.swing.*;
import java.io.*;
import java.net.*;

public class AWTUtils {
    public static Dimension screenSize() {
	return Toolkit.getDefaultToolkit().getScreenSize();
    }

    public static int screenWidth() {
	return (int)screenSize().getWidth();
    }

    public static int screenHeight() {
	return (int)screenSize().getHeight();
    }

    public static void placeAtTheCenter(Window frame) {
	frame.setLocation((int)(screenWidth()/2 - frame.getWidth()/2), (int)((screenHeight()/2 - frame.getHeight()/2)));
    }

    public static void maximize(Frame frame) {
	frame.setExtendedState(Frame.MAXIMIZED_BOTH);
    }

    public static void forceSize(Component c, Dimension d) {
	c.setMinimumSize(d);
	c.setMaximumSize(d);
	c.setPreferredSize(d);
    }

    public static void forceSize(Component c, int w, int h) { forceSize(c, new Dimension(w, h)); }

    public static Dimension getContentSize(Container c) {
	Dimension size = c.getSize();
	Insets insets = c.getInsets();

	return new Dimension(size.width - insets.left - insets.right, size.height - insets.top - insets.bottom);
    }

    public static void setContentSize(JFrame f, Dimension d) {
	Insets insets = f.getInsets();
	f.setSize(d.width + insets.left + insets.right, d.height + insets.top + insets.bottom);
    }

    public static void setContentSize(JFrame f, int w, int h) {
	Insets insets = f.getInsets();
	f.setSize(w + insets.left + insets.right, h + insets.top + insets.bottom);
    }

    public static void makeFocusable(final Component c) {
	c.setFocusable(true);
	c.addMouseListener(new MouseListener() {
		public void mousePressed(MouseEvent e) {}
		public void mouseReleased(MouseEvent e) {}
		public void mouseEntered(MouseEvent e) {}
		public void mouseExited(MouseEvent e) {}
		public void mouseClicked(MouseEvent e) {
		    c.requestFocusInWindow();
		}
	    });
    }

    public static void bringToTheFront(Window w) {
	w.setAlwaysOnTop(true);
	w.setAlwaysOnTop(false);
    }

    public static void bringToTheFront(Frame frame) {
	frame.setAlwaysOnTop(true);
	frame.setAlwaysOnTop(false);
	frame.setExtendedState(Frame.NORMAL);
	frame.requestFocus();
    }

    public static BufferedImage captureScreen() {
	BufferedImage img = null;

	try {
	    img = new Robot().createScreenCapture(new Rectangle(screenSize()));
	} catch(AWTException e) {
	}

	return img;
    }

    public static void enableAWTHeadless(boolean b) {
	System.setProperty("java.awt.headless", b ? "true" : "false");
    }

    public static boolean setNativeLookAndFeel() {
	boolean ret = true;

	try {
	    UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
	} catch(Exception e) {
	    ret = false;
	}

	return ret;
    }

    public static String showSaveDialog(String lastPath, FilenameFilter filter) {
	Frame frame = null;

	FileDialog fd = new java.awt.FileDialog(frame);

	if (lastPath != null) fd.setDirectory(lastPath);
	else fd.setDirectory((String)System.getProperties().get("user.dir"));

	fd.setFilenameFilter(filter);

	AWTUtils.placeAtTheCenter(fd);

	fd.setMode(FileDialog.SAVE);
	fd.setVisible(true);

	if (fd.getFile() == null) return null;

	String fn = fd.getDirectory() + fd.getFile();

	return fn;
    }

    public static String showSaveDialog(String lastPath, String regex) {
	return showSaveDialog(lastPath, new RegexFileFilter(regex, ""));
    }

    public static String showLoadDialog(String lastPath, FilenameFilter filter) {
	Frame frame = null;

	FileDialog fd = new java.awt.FileDialog(frame);

	if (lastPath != null) fd.setDirectory(lastPath);
	else fd.setDirectory((String)System.getProperties().get("user.dir"));

	fd.setFilenameFilter(filter);

	AWTUtils.placeAtTheCenter(fd);

	fd.setMode(FileDialog.LOAD);
	fd.setVisible(true);

	if (fd.getFile() == null) return null;

	String fn = fd.getDirectory() + fd.getFile();

	return fn;
    }

    public static String showLoadDialog(String lastPath, String regex) {
	return showLoadDialog(lastPath, new RegexFileFilter(regex, ""));
    }

    public static void openURL(String url) {
	Desktop desktop = Desktop.getDesktop();

        if (!desktop.isSupported(Desktop.Action.BROWSE)) return;

	try {
	    URI uri = new URI(url);
	    desktop.browse(uri);
	} catch (Exception e) {
	    System.err.println( e.getMessage() );
	}
    }
}
