package org.naokishibata.generic;

import java.io.*;
import java.util.*;
import javax.swing.*;

public class Debug {
    public static boolean abortViaGUI = false;

    public static void setAbortViaGUI(boolean b) { abortViaGUI = b; }

    public static final void abort(String s, Exception e) {
	if (abortViaGUI) abortGUI(s, e);

	System.err.println("Aborted : " + s);
	e.printStackTrace();
	System.exit(-1);
    }

    public static final void abort(String s) {
	if (abortViaGUI) abortGUI(s);

	System.err.println("Aborted(" + (new GregorianCalendar()).getTime()+ ") : " + s);
	Thread.dumpStack();
	System.exit(-1);
    }

    public static final void abort(Exception e) {
	if (abortViaGUI) abortGUI(e);

	System.err.println("Aborted :");
	e.printStackTrace();
	System.exit(-1);
    }

    public static void abort() {
	if (abortViaGUI) abortGUI("");

	System.err.println("Aborted(" + (new GregorianCalendar()).getTime()+ ")");
	Thread.dumpStack();
	System.exit(-1);
    }

    public static void abortGUI(String s) {
	abortGUI(s, null);
    }

    public static void abortGUI(Throwable e) {
	abortGUI("", e);
    }

    public static String throwableToString(Throwable e) {
	String s = e + "\n";

	StackTraceElement[] a = e.getStackTrace();

	for(int i=0;i<a.length;i++) s = s + a[i] + "\n";

	return s;
    }

    public static void abortGUI(String s, Throwable e) {
	if (e != null) {
	    s = s + "\n" + e + "\n";

	    StackTraceElement[] a = e.getStackTrace();

	    for(int i=0;i<a.length;i++) s = s + a[i] + "\n";
	} else {
	    s = s + "\n";

	    StackTraceElement[] a = Thread.currentThread().getStackTrace();

	    for(int i=0;i<a.length;i++) s = s + a[i] + "\n";
	}

        JOptionPane.showMessageDialog(null, s, "Program aborting", JOptionPane.ERROR_MESSAGE);

        System.exit(-1);
    }

    public static void showException(String s, Throwable e) {
	s = s + "\n" + e + "\n";

	StackTraceElement[] a = e.getStackTrace();

	for(int i=0;i<a.length;i++) s = s + a[i] + "\n";

        JOptionPane.showMessageDialog(null, s, "Exception", JOptionPane.ERROR_MESSAGE);
    }

    static class UEHandlerThreadGroup extends ThreadGroup {
	UEHandlerThreadGroup() { super("UEHandlerThreadGroup"); }

	public void uncaughtException(Thread t, Throwable e) {
	    abortGUI("Uncaught Exception from thread " + t.getName(), e);
	}
    }

    static public void showAllUncaughtExceptionViaGUI() {
	Thread.currentThread().setUncaughtExceptionHandler(new UEHandlerThreadGroup());
    }
}

